@@ -49,9 +49,11 @@
 #define CA_TAG "ca "
 #define CERT_TAG "cert "
 #define CIPHER_TAG "cipher "
+#define KEYSIZE_TAG "keysize "
 #define CLIENT_TAG "client"
 #define COMP_TAG "comp-lzo"
 #define DEV_TAG "dev "
+#define DEV_TYPE_TAG "dev-type "
 #define FRAGMENT_TAG "fragment "
 #define IFCONFIG_TAG "ifconfig "
 #define KEY_TAG "key "
@@ -324,12 +326,21 @@ do_import (const char *path, char **lines, GError **error)
 		if (!strncmp (*line, DEV_TAG, strlen (DEV_TAG))) {
 			items = get_args (*line + strlen (DEV_TAG), &nitems);
 			if (nitems == 1) {
-				if (g_str_has_prefix (items[0], "tun")) {
-					/* ignore; default is tun */
-				} else if (g_str_has_prefix (items[0], "tap"))
-					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_TAP_DEV, "yes");
+				nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_DEV, items[0]);
+			} else
+				g_warning ("%s: invalid number of arguments in option '%s'", __func__, *line);
+
+			g_strfreev (items);
+			continue;
+		}
+
+		if (!strncmp (*line, DEV_TYPE_TAG, strlen (DEV_TYPE_TAG))) {
+			items = get_args (*line + strlen (DEV_TYPE_TAG), &nitems);
+			if (nitems == 1) {
+				if (!strcmp (items[0], "tun") || !strcmp (items[0], "tap"))
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_DEV_TYPE, items[0]);
 				else
-					g_warning ("%s: unknown %s option '%s'", __func__, DEV_TAG, *line);
+					g_warning ("%s: unknown %s option '%s'", __func__, DEV_TYPE_TAG, *line);
 			} else
 				g_warning ("%s: invalid number of arguments in option '%s'", __func__, *line);
 
@@ -431,9 +442,7 @@ do_import (const char *path, char **lines, GError **error)
 
 		if (   !strncmp (*line, HTTP_PROXY_RETRY_TAG, strlen (HTTP_PROXY_RETRY_TAG))
 		    || !strncmp (*line, SOCKS_PROXY_RETRY_TAG, strlen (SOCKS_PROXY_RETRY_TAG))) {
-			nm_setting_vpn_add_data_item (s_vpn,
-			                              g_strdup (NM_OPENVPN_KEY_PROXY_RETRY),
-			                              g_strdup ("yes"));
+			nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_RETRY, "yes");
 			continue;
 		}
 
@@ -597,6 +606,25 @@ do_import (const char *path, char **lines, GError **error)
 			continue;
 		}
 
+		if (!strncmp (*line, KEYSIZE_TAG, strlen (KEYSIZE_TAG))) {
+			items = get_args (*line + strlen (KEYSIZE_TAG), &nitems);
+			if (nitems == 1) {
+				glong key_size;
+
+				errno = 0;
+				key_size = strtol (items[0], NULL, 10);
+				if ((errno == 0) && (key_size > 0) && (key_size <= 65535)) {
+					tmp = g_strdup_printf ("%d", (guint32) key_size);
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_KEYSIZE, tmp);
+					g_free (tmp);
+				} else
+					g_warning ("%s: invalid key size in option '%s'", __func__, *line);
+			} else
+				g_warning ("%s: invalid number of arguments in option '%s'", __func__, *line);
+			g_strfreev (items);
+			continue;
+		}
+
 		if (!strncmp (*line, TLS_REMOTE_TAG, strlen (TLS_REMOTE_TAG))) {
 			char *unquoted = unquote (*line + strlen (TLS_REMOTE_TAG), NULL);
 
@@ -748,12 +776,16 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	const char *remote_cert_tls = NULL;
 	const char *tls_auth = NULL;
 	const char *tls_auth_dir = NULL;
+	const char *device = NULL;
+	const char *device_type = NULL;
+	const char *device_default = "tun";
 	gboolean success = FALSE;
-	gboolean device_tun = TRUE;
 	gboolean proto_udp = TRUE;
 	gboolean use_lzo = FALSE;
 	gboolean reneg_exists = FALSE;
 	guint32 reneg = 0;
+	gboolean keysize_exists = FALSE;
+	guint32 keysize = 0;
 	const char *proxy_type = NULL;
 	const char *proxy_server = NULL;
 	const char *proxy_port = NULL;
@@ -833,9 +865,18 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (value && !strcmp (value, "yes"))
 		proto_udp = FALSE;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV);
+	if (value && strlen (value))
+		device = value;
+
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV_TYPE);
+	if (value && strlen (value))
+		device_type = value;
+
+	/* Read legacy 'tap-dev' property for backwards compatibility. */
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TAP_DEV);
 	if (value && !strcmp (value, "yes"))
-		device_tun = FALSE;
+		device_default = "tap";
 
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMP_LZO);
 	if (value && !strcmp (value, "yes"))
@@ -845,6 +886,12 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (value && strlen (value))
 		cipher = value;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEYSIZE);
+	if (value && strlen (value)) {
+		keysize_exists = TRUE;
+		keysize = strtol (value, NULL, 10);
+	}
+
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_LOCAL_IP);
 	if (value && strlen (value))
 		local_ip = value;
@@ -913,6 +960,9 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (cipher)
 		fprintf (f, "cipher %s\n", cipher);
 
+	if (keysize_exists)
+		fprintf (f, "keysize %d\n", keysize);
+
 	if (use_lzo)
 		fprintf (f, "comp-lzo yes\n");
 
@@ -930,7 +980,9 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	if (value && strlen (value))
 		fprintf (f, FRAGMENT_TAG " %d\n", (int) strtol (value, NULL, 10));
 
-	fprintf (f, "dev %s\n", device_tun ? "tun" : "tap");
+	fprintf (f, "dev %s\n", device ? device : (device_type ? device_type : device_default));
+	if (device_type)
+		fprintf (f, "dev-type %s\n", device_type);
 	fprintf (f, "proto %s\n", proto_udp ? "udp" : "tcp");
 
 	if (local_ip && remote_ip)
