@@ -692,7 +692,9 @@ get_suggested_name (NMVpnPluginUiInterface *iface, NMConnection *connection)
 static guint32
 get_capabilities (NMVpnPluginUiInterface *iface)
 {
-	return (NM_VPN_PLUGIN_UI_CAPABILITY_IMPORT | NM_VPN_PLUGIN_UI_CAPABILITY_EXPORT);
+	return (NM_VPN_PLUGIN_UI_CAPABILITY_IMPORT |
+	        NM_VPN_PLUGIN_UI_CAPABILITY_EXPORT |
+	        NM_VPN_PLUGIN_UI_CAPABILITY_IPV6);
 }
 
 static NMVpnPluginUiWidgetInterface *
@@ -764,6 +766,9 @@ nm_vpn_plugin_ui_factory (GError **error)
 	if (error)
 		g_return_val_if_fail (*error == NULL, NULL);
 
+	bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
+	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+
 	return NM_VPN_PLUGIN_UI_INTERFACE (g_object_new (OPENVPN_TYPE_PLUGIN_UI, NULL));
 }
 
